@@ -1,40 +1,45 @@
 package hyperconverged
 
 import (
+	"context"
 	"fmt"
 	apiextensionsv1 "k8s.io/apiextensions-apiserver/pkg/apis/apiextensions/v1"
 	"os"
+	"reflect"
+	"strings"
 
 	"github.com/google/uuid"
 	conditionsv1 "github.com/openshift/custom-resource-status/conditions/v1"
 	operatorhandler "github.com/operator-framework/operator-lib/handler"
-
 	corev1 "k8s.io/api/core/v1"
 	schedulingv1 "k8s.io/api/scheduling/v1"
 	apierrors "k8s.io/apimachinery/pkg/api/errors"
+	metav1 "k8s.io/apimachinery/pkg/apis/meta/v1"
 	"k8s.io/apimachinery/pkg/runtime"
 	"k8s.io/apimachinery/pkg/types"
+	"k8s.io/apimachinery/pkg/util/yaml"
 	"k8s.io/client-go/tools/record"
 	"sigs.k8s.io/controller-runtime/pkg/client"
 	"sigs.k8s.io/controller-runtime/pkg/controller"
 	"sigs.k8s.io/controller-runtime/pkg/handler"
+	logf "sigs.k8s.io/controller-runtime/pkg/log"
 	"sigs.k8s.io/controller-runtime/pkg/manager"
+	"sigs.k8s.io/controller-runtime/pkg/predicate"
 	"sigs.k8s.io/controller-runtime/pkg/reconcile"
-	logf "sigs.k8s.io/controller-runtime/pkg/runtime/log"
 	"sigs.k8s.io/controller-runtime/pkg/source"
 
-	monitoringv1 "github.com/coreos/prometheus-operator/pkg/apis/monitoring/v1"
 	networkaddonsv1 "github.com/kubevirt/cluster-network-addons-operator/pkg/apis/networkaddonsoperator/v1"
+	vmimportv1beta1 "github.com/kubevirt/vm-import-operator/pkg/apis/v2v/v1beta1"
+	monitoringv1 "github.com/prometheus-operator/prometheus-operator/pkg/apis/monitoring/v1"
+	kubevirtv1 "kubevirt.io/client-go/api/v1"
+	cdiv1beta1 "kubevirt.io/containerized-data-importer/pkg/apis/core/v1beta1"
+	sspv1beta1 "kubevirt.io/ssp-operator/api/v1beta1"
+
 	hcov1beta1 "github.com/kubevirt/hyperconverged-cluster-operator/pkg/apis/hco/v1beta1"
 	"github.com/kubevirt/hyperconverged-cluster-operator/pkg/controller/common"
 	"github.com/kubevirt/hyperconverged-cluster-operator/pkg/controller/operands"
 	hcoutil "github.com/kubevirt/hyperconverged-cluster-operator/pkg/util"
-	"github.com/kubevirt/hyperconverged-cluster-operator/pkg/util/predicate"
 	version "github.com/kubevirt/hyperconverged-cluster-operator/version"
-	vmimportv1beta1 "github.com/kubevirt/vm-import-operator/pkg/apis/v2v/v1beta1"
-	kubevirtv1 "kubevirt.io/client-go/api/v1"
-	cdiv1beta1 "kubevirt.io/containerized-data-importer/pkg/apis/core/v1beta1"
-	sspv1beta1 "kubevirt.io/ssp-operator/api/v1beta1"
 )
 
 var (
@@ -65,7 +70,7 @@ const (
 	secondaryCRPrefix = "hco-controlled-cr-"
 )
 
-// Annotations used to patch operand CRs with unsupported/unofficial/hidden features.
+// JSONPatchAnnotationNames - annotations used to patch operand CRs with unsupported/unofficial/hidden features.
 // The presence of any of these annotations raises the hcov1beta1.ConditionTaintedConfiguration condition.
 var JSONPatchAnnotationNames = []string{
 	common.JSONPatchKVAnnotationName,
@@ -73,9 +78,8 @@ var JSONPatchAnnotationNames = []string{
 	common.JSONPatchCNAOAnnotationName,
 }
 
-// Add creates a new HyperConverged Controller and adds it to the Manager. The Manager will set fields on the Controller
-// and Start it when the Manager is Started.
-func Add(mgr manager.Manager, ci hcoutil.ClusterInfo) error {
+// RegisterReconciler creates a new HyperConverged Reconciler and registers it into manager.
+func RegisterReconciler(mgr manager.Manager, ci hcoutil.ClusterInfo) error {
 	return add(mgr, newReconciler(mgr, ci), ci)
 }
 
@@ -88,15 +92,14 @@ func newReconciler(mgr manager.Manager, ci hcoutil.ClusterInfo) reconcile.Reconc
 	}
 
 	return &ReconcileHyperConverged{
-		client:             mgr.GetClient(),
-		scheme:             mgr.GetScheme(),
-		recorder:           mgr.GetEventRecorderFor(hcoutil.HyperConvergedName),
-		cliDownloadHandler: &operands.CLIDownloadHandler{Client: mgr.GetClient(), Scheme: mgr.GetScheme()},
-		operandHandler:     operands.NewOperandHandler(mgr.GetClient(), mgr.GetScheme(), ci.IsOpenshift(), hcoutil.GetEventEmitter()),
-		upgradeMode:        false,
-		ownVersion:         ownVersion,
-		eventEmitter:       hcoutil.GetEventEmitter(),
-		firstLoop:          true,
+		client:         mgr.GetClient(),
+		scheme:         mgr.GetScheme(),
+		recorder:       mgr.GetEventRecorderFor(hcoutil.HyperConvergedName),
+		operandHandler: operands.NewOperandHandler(mgr.GetClient(), mgr.GetScheme(), ci.IsOpenshift(), hcoutil.GetEventEmitter()),
+		upgradeMode:    false,
+		ownVersion:     ownVersion,
+		eventEmitter:   hcoutil.GetEventEmitter(),
+		firstLoop:      true,
 	}
 }
 
@@ -112,7 +115,7 @@ func add(mgr manager.Manager, r reconcile.Reconciler, ci hcoutil.ClusterInfo) er
 	err = c.Watch(
 		&source.Kind{Type: &hcov1beta1.HyperConverged{}},
 		&operatorhandler.InstrumentedEnqueueRequestForObject{},
-		predicate.GenerationOrAnnotationChangedPredicate{})
+		predicate.Or(predicate.GenerationChangedPredicate{}, predicate.AnnotationChangedPredicate{}))
 	if err != nil {
 		return err
 	}
@@ -122,7 +125,7 @@ func add(mgr manager.Manager, r reconcile.Reconciler, ci hcoutil.ClusterInfo) er
 		return err
 	}
 
-	secondaryResources := []runtime.Object{
+	secondaryResources := []client.Object{
 		&kubevirtv1.KubeVirt{},
 		&cdiv1beta1.CDI{},
 		&networkaddonsv1.NetworkAddonsConfig{},
@@ -131,7 +134,7 @@ func add(mgr manager.Manager, r reconcile.Reconciler, ci hcoutil.ClusterInfo) er
 		&vmimportv1beta1.VMImportConfig{},
 	}
 	if ci.IsOpenshift() {
-		secondaryResources = append(secondaryResources, []runtime.Object{
+		secondaryResources = append(secondaryResources, []client.Object{
 			&corev1.Service{},
 			&monitoringv1.ServiceMonitor{},
 			&monitoringv1.PrometheusRule{},
@@ -141,18 +144,18 @@ func add(mgr manager.Manager, r reconcile.Reconciler, ci hcoutil.ClusterInfo) er
 	// Watch secondary resources
 	for _, resource := range secondaryResources {
 		msg := fmt.Sprintf("Reconciling for %T", resource)
-		err = c.Watch(&source.Kind{Type: resource}, &handler.EnqueueRequestsFromMapFunc{
-			ToRequests: handler.ToRequestsFunc(
+		err = c.Watch(
+			&source.Kind{Type: resource},
+			handler.EnqueueRequestsFromMapFunc(func(a client.Object) []reconcile.Request {
 				// enqueue using a placeholder to be able to discriminate request triggered
 				// by changes on the HyperConverged object from request triggered by changes
 				// on a secondary CR controlled by HCO
-				func(a handler.MapObject) []reconcile.Request {
-					log.Info(msg)
-					return []reconcile.Request{
-						{NamespacedName: secCRPlaceholder},
-					}
-				}),
-		})
+				log.Info(msg)
+				return []reconcile.Request{
+					{NamespacedName: secCRPlaceholder},
+				}
+			}),
+		)
 		if err != nil {
 			return err
 		}
@@ -167,15 +170,14 @@ var _ reconcile.Reconciler = &ReconcileHyperConverged{}
 type ReconcileHyperConverged struct {
 	// This client, initialized using mgr.Client() above, is a split client
 	// that reads objects from the cache and writes to the apiserver
-	client             client.Client
-	scheme             *runtime.Scheme
-	recorder           record.EventRecorder
-	cliDownloadHandler *operands.CLIDownloadHandler
-	operandHandler     *operands.OperandHandler
-	upgradeMode        bool
-	ownVersion         string
-	eventEmitter       hcoutil.EventEmitter
-	firstLoop          bool
+	client         client.Client
+	scheme         *runtime.Scheme
+	recorder       record.EventRecorder
+	operandHandler *operands.OperandHandler
+	upgradeMode    bool
+	ownVersion     string
+	eventEmitter   hcoutil.EventEmitter
+	firstLoop      bool
 }
 
 // Reconcile reads that state of the cluster for a HyperConverged object and makes changes based on the state read
@@ -183,104 +185,79 @@ type ReconcileHyperConverged struct {
 // Note:
 // The Controller will requeue the Request to be processed again if the returned error is non-nil or
 // Result.Requeue is true, otherwise upon completion it will remove the work from the queue.
-func (r *ReconcileHyperConverged) Reconcile(request reconcile.Request) (reconcile.Result, error) {
+func (r *ReconcileHyperConverged) Reconcile(ctx context.Context, request reconcile.Request) (reconcile.Result, error) {
+	logger := log.WithValues("Request.Namespace", request.Namespace, "Request.Name", request.Name)
 
-	secCRPlaceholder, err := getSecondaryCRPlaceholder()
+	hcoTriggered, err := isTriggeredByHyperConverged(request)
 	if err != nil {
 		return reconcile.Result{}, err
 	}
 
-	hcoTriggered := true
-	pRequest := request
-	if request.NamespacedName == secCRPlaceholder {
-		hcoTriggered = false
-		hco, err := getHyperconverged()
-		if err != nil {
-			return reconcile.Result{}, err
-		}
-		pRequest = reconcile.Request{
-			NamespacedName: hco,
-		}
+	resolvedRequest, err := resolveReconcileRequest(request, hcoTriggered)
+	if err != nil {
+		return reconcile.Result{}, err
 	}
+	hcoRequest := common.NewHcoRequest(ctx, resolvedRequest, log, r.upgradeMode, hcoTriggered)
 
-	req := common.NewHcoRequest(pRequest, log, r.upgradeMode, hcoTriggered)
-	if req.HCOTriggered {
-		req.Logger.Info("Reconciling HyperConverged operator")
+	if hcoTriggered {
+		logger.Info("Reconciling HyperConverged operator")
+		r.operandHandler.Reset()
 	} else {
-		req.Logger.Info("The reconciliation got triggered by a secondary CR object")
+		logger.Info("The reconciliation got triggered by a secondary CR object")
 	}
 
 	// Fetch the HyperConverged instance
-	instance, err := r.getHcoInstanceFromK8s(req)
+	instance, err := r.getHyperConverged(hcoRequest)
 	if instance == nil {
 		return reconcile.Result{}, err
 	}
-	req.Instance = instance
+	hcoRequest.Instance = instance
 
 	if r.firstLoop {
-		// reload eventEmitter. The client should now find all the required resources
-		r.eventEmitter.UpdateClient(req.Ctx, r.client, req.Logger)
-
-		r.operandHandler.FirstUseInitiation(r.scheme, hcoutil.GetClusterInfo().IsOpenshift(), req.Instance)
-	}
-
-	res, err := r.doReconcile(req)
-	if r.firstLoop {
-		r.firstLoop = false
+		r.firstLoopInitialization(hcoRequest)
 	}
 
+	result, err := r.doReconcile(hcoRequest)
 	if err != nil {
-		r.eventEmitter.EmitEvent(req.Instance, corev1.EventTypeWarning, "ReconcileError", err.Error())
-	}
-
-	/*
-		From K8s API reference: https://kubernetes.io/docs/reference/generated/kubernetes-api/v1.18/
-		============================================================================================================
-		Replace: Replacing a resource object will update the resource by replacing the existing spec with the
-		provided one. For read-then-write operations this is safe because an optimistic lock failure will occur if
-		the resource was modified between the read and write.
-
-		**Note: The ResourceStatus will be ignored by the system and will not be updated. To update the status, one
-		must invoke the specific status update operation.**
-		============================================================================================================
-
-		In addition, updating the status should not update the metadata, so we need to update both the CR and the
-		CR Status, and we need to update the status first, in order to prevent a conflict.
-	*/
-
-	if req.StatusDirty {
-		updateErr := r.client.Status().Update(req.Ctx, req.Instance)
-		if updateErr != nil {
-			updateErrorMsg := "Failed to update HCO Status"
-			r.eventEmitter.EmitEvent(req.Instance, corev1.EventTypeWarning, "HcoUpdateError", updateErrorMsg)
-			req.Logger.Error(updateErr, updateErrorMsg)
-			err = updateErr
-		}
-	}
-
-	// recover Spec.Version if upgrade missed when upgrade completed
-	// Doing it here because status.update overrides spec for some reason
-	knownHcoVersion, versionFound := req.Instance.Status.GetVersion(hcoVersionName)
-	if (!r.upgradeMode) && versionFound && (knownHcoVersion == r.ownVersion) && (req.Instance.Spec.Version != r.ownVersion) {
-		req.Instance.Spec.Version = r.ownVersion
-		req.Dirty = true
-	}
-
-	if req.Dirty {
-		updateErr := r.client.Update(req.Ctx, req.Instance)
-		if updateErr != nil {
-			updateErrorMsg := "Failed to update HCO CR"
-			r.eventEmitter.EmitEvent(req.Instance, corev1.EventTypeWarning, "HcoUpdateError", updateErrorMsg)
-			req.Logger.Error(updateErr, updateErrorMsg)
-			err = updateErr
-		}
+		r.eventEmitter.EmitEvent(hcoRequest.Instance, corev1.EventTypeWarning, "ReconcileError", err.Error())
+		return reconcile.Result{}, err
 	}
 
+	err = r.updateHyperConverged(hcoRequest)
 	if apierrors.IsConflict(err) {
-		res.Requeue = true
+		result.Requeue = true
 	}
 
-	return res, err
+	return result, err
+}
+
+// resolveReconcileRequest returns a reconcile.Request to be used throughout the reconciliation cycle,
+// regardless of which resource has triggered it.
+func resolveReconcileRequest(originalRequest reconcile.Request, hcoTriggered bool) (reconcile.Request, error) {
+	if hcoTriggered {
+		return originalRequest, nil
+	}
+
+	hc, err := getHyperConvergedNamespacedName()
+	if err != nil {
+		return reconcile.Request{}, err
+	}
+
+	resolvedRequest := reconcile.Request{
+		NamespacedName: hc,
+	}
+
+	return resolvedRequest, nil
+}
+
+func isTriggeredByHyperConverged(request reconcile.Request) (bool, error) {
+	placeholder, err := getSecondaryCRPlaceholder()
+	if err != nil {
+		return false, err
+	}
+
+	isHyperConverged := request.NamespacedName != placeholder
+	return isHyperConverged, nil
 }
 
 func (r *ReconcileHyperConverged) doReconcile(req *common.HcoRequest) (reconcile.Result, error) {
@@ -302,21 +279,8 @@ func (r *ReconcileHyperConverged) doReconcile(req *common.HcoRequest) (reconcile
 	// negative conditions (!Available, Degraded, Progressing)
 	req.Conditions = common.NewHcoConditions()
 
-	finDropped := false
 	// Handle finalizers
-	if contains(req.Instance.ObjectMeta.Finalizers, badFinalizerName) {
-		req.Logger.Info("removing a finalizer set in the past (without a fully qualified name)")
-		req.Instance.ObjectMeta.Finalizers, finDropped = drop(req.Instance.ObjectMeta.Finalizers, badFinalizerName)
-		req.Dirty = req.Dirty || finDropped
-	}
-	if req.Instance.ObjectMeta.DeletionTimestamp.IsZero() {
-		// Add the finalizer if it's not there
-		if !contains(req.Instance.ObjectMeta.Finalizers, FinalizerName) {
-			req.Logger.Info("setting a finalizer (with fully qualified name)")
-			req.Instance.ObjectMeta.Finalizers = append(req.Instance.ObjectMeta.Finalizers, FinalizerName)
-			req.Dirty = req.Dirty || finDropped
-		}
-	} else {
+	if !checkFinalizers(req) {
 		if !req.HCOTriggered {
 			// this is just the effect of a delete request created by HCO
 			// in the previous iteration, ignore it
@@ -331,8 +295,14 @@ func (r *ReconcileHyperConverged) doReconcile(req *common.HcoRequest) (reconcile
 
 	if !r.upgradeMode && !init && knownHcoVersion != r.ownVersion {
 		r.upgradeMode = true
+
 		r.eventEmitter.EmitEvent(req.Instance, corev1.EventTypeNormal, "UpgradeHCO", "Upgrading the HyperConverged to version "+r.ownVersion)
 		req.Logger.Info(fmt.Sprintf("Start upgrading from version %s to version %s", knownHcoVersion, r.ownVersion))
+	}
+
+	req.SetUpgradeMode(r.upgradeMode)
+
+	if r.upgradeMode {
 		crdStatusUpdated, err := r.updateCrdStoredVersions(req)
 		if err != nil {
 			return reconcile.Result{Requeue: true}, err
@@ -341,14 +311,23 @@ func (r *ReconcileHyperConverged) doReconcile(req *common.HcoRequest) (reconcile
 				return reconcile.Result{Requeue: true}, nil
 			}
 		}
+
+		modified, err := r.migrateBeforeUpgrade(req)
+		if err != nil {
+			return reconcile.Result{Requeue: init}, err
+		}
+
+		if modified {
+			r.updateConditions(req)
+			return reconcile.Result{Requeue: true}, nil
+		}
 	}
 
-	req.SetUpgradeMode(r.upgradeMode)
+	return r.EnsureOperandAndComplete(req, init)
+}
 
-	r.cliDownloadHandler.Ensure(req)
-
-	err = r.operandHandler.Ensure(req)
-	if err != nil {
+func (r *ReconcileHyperConverged) EnsureOperandAndComplete(req *common.HcoRequest, init bool) (reconcile.Result, error) {
+	if err := r.operandHandler.Ensure(req); err != nil {
 		r.updateConditions(req)
 		hcoutil.SetReady(false)
 		return reconcile.Result{Requeue: init}, nil
@@ -359,7 +338,7 @@ func (r *ReconcileHyperConverged) doReconcile(req *common.HcoRequest) (reconcile
 	// Requeue if we just created everything
 	if init {
 		hcoutil.SetReady(false)
-		return reconcile.Result{Requeue: true}, err
+		return reconcile.Result{Requeue: true}, nil
 	}
 
 	r.completeReconciliation(req)
@@ -367,25 +346,98 @@ func (r *ReconcileHyperConverged) doReconcile(req *common.HcoRequest) (reconcile
 	return reconcile.Result{}, nil
 }
 
-func (r *ReconcileHyperConverged) getHcoInstanceFromK8s(req *common.HcoRequest) (*hcov1beta1.HyperConverged, error) {
+// getHyperConverged gets the HyperConverged resource from the Kubernetes API.
+func (r *ReconcileHyperConverged) getHyperConverged(req *common.HcoRequest) (*hcov1beta1.HyperConverged, error) {
 	instance := &hcov1beta1.HyperConverged{}
 	err := r.client.Get(req.Ctx, req.NamespacedName, instance)
-	if err != nil {
-		if apierrors.IsNotFound(err) {
-			req.Logger.Info("No HyperConverged resource")
-			// Request object not found, could have been deleted after reconcile request.
-			// Owned objects are automatically garbage collected. For additional cleanup logic use finalizers.
-			// Return and don't requeue
-			return nil, nil
-		}
-		// Error reading the object - requeue the request.
-		return nil, err
+
+	// Green path first
+	if err == nil {
+		return instance, nil
 	}
-	return instance, nil
+
+	// Error path
+	if apierrors.IsNotFound(err) {
+		req.Logger.Info("No HyperConverged resource")
+		// Request object not found, could have been deleted after reconcile request.
+		// Owned objects are automatically garbage collected. For additional cleanup logic use finalizers.
+		// Return and don't requeue
+		return nil, nil
+	}
+
+	// Another error reading the object.
+	// Just return the error so that the request is requeued.
+	return nil, err
+}
+
+// updateHyperConverged updates the HyperConverged resource according to its state in the request.
+func (r *ReconcileHyperConverged) updateHyperConverged(request *common.HcoRequest) error {
+
+	// Since the status subresource is enabled for the HyperConverged kind,
+	// we need to update the status and the metadata separately.
+	// Moreover, we need to update the status first, in order to prevent a conflict.
+	// In addition, spec changes are removed by status update, but since status update done first, we need
+	// to store the spec and recover it after status update
+
+	var spec hcov1beta1.HyperConvergedSpec
+	if request.Dirty {
+		request.Instance.Spec.DeepCopyInto(&spec)
+	}
+
+	err := r.updateHyperConvergedStatus(request)
+	if err != nil {
+		r.logHyperConvergedUpdateError(request, err, "Failed to update HCO Status")
+		return err
+	}
+
+	// restore spec
+	if request.Dirty {
+		request.Instance.Spec = spec
+	}
+
+	// Doing it here because status.update overrides spec for some reason
+	r.recoverHCOVersion(request)
+
+	err = r.updateHyperConvergedSpecMetadata(request)
+	if err != nil {
+		r.logHyperConvergedUpdateError(request, err, "Failed to update HCO CR")
+		return err
+	}
+
+	return nil
+}
+
+// updateHyperConvergedSpecMetadata updates the HyperConverged resource's spec and metadata.
+func (r *ReconcileHyperConverged) updateHyperConvergedSpecMetadata(request *common.HcoRequest) error {
+	if !request.Dirty {
+		return nil
+	}
+
+	return r.client.Update(request.Ctx, request.Instance)
+}
+
+// updateHyperConvergedSpecMetadata updates the HyperConverged resource's status (and metadata).
+func (r *ReconcileHyperConverged) updateHyperConvergedStatus(request *common.HcoRequest) error {
+	if !request.StatusDirty {
+		return nil
+	}
+
+	return r.client.Status().Update(request.Ctx, request.Instance)
+}
+
+// logHyperConvergedUpdateError logs an error that occurred during resource update,
+// as well as emits a corresponding event.
+func (r *ReconcileHyperConverged) logHyperConvergedUpdateError(request *common.HcoRequest, err error, errMsg string) {
+	r.eventEmitter.EmitEvent(request.Instance,
+		corev1.EventTypeWarning,
+		"HcoUpdateError",
+		errMsg)
+
+	request.Logger.Error(err, errMsg)
 }
 
 func (r *ReconcileHyperConverged) validateNamespace(req *common.HcoRequest) (bool, error) {
-	hco, err := getHyperconverged()
+	hco, err := getHyperConvergedNamespacedName()
 	if err != nil {
 		req.Logger.Error(err, "Failed to get HyperConverged namespaced name")
 		return false, err
@@ -454,16 +506,16 @@ func (r *ReconcileHyperConverged) ensureHcoDeleted(req *common.HcoRequest) (reco
 	requeue := false
 
 	// Remove the finalizers
-	fin_dropped := false
-	if contains(req.Instance.ObjectMeta.Finalizers, FinalizerName) {
-		req.Instance.ObjectMeta.Finalizers, fin_dropped = drop(req.Instance.ObjectMeta.Finalizers, FinalizerName)
+	finDropped := false
+	if hcoutil.ContainsString(req.Instance.ObjectMeta.Finalizers, FinalizerName) {
+		req.Instance.ObjectMeta.Finalizers, finDropped = drop(req.Instance.ObjectMeta.Finalizers, FinalizerName)
 		req.Dirty = true
-		requeue = requeue || fin_dropped
+		requeue = requeue || finDropped
 	}
-	if contains(req.Instance.ObjectMeta.Finalizers, badFinalizerName) {
-		req.Instance.ObjectMeta.Finalizers, fin_dropped = drop(req.Instance.ObjectMeta.Finalizers, badFinalizerName)
+	if hcoutil.ContainsString(req.Instance.ObjectMeta.Finalizers, badFinalizerName) {
+		req.Instance.ObjectMeta.Finalizers, finDropped = drop(req.Instance.ObjectMeta.Finalizers, badFinalizerName)
 		req.Dirty = true
-		requeue = requeue || fin_dropped
+		requeue = requeue || finDropped
 	}
 
 	// Need to requeue because finalizer update does not change metadata.generation
@@ -578,7 +630,7 @@ func (r *ReconcileHyperConverged) aggregateComponentConditions(req *common.HcoRe
 			captured at one time (ie. if KubeVirt and CDI are both reporting !Available,
 		    you will only see CDI as it updates last).
 	*/
-	allComponentsAreUp := req.Conditions.Empty()
+	allComponentsAreUp := req.Conditions.IsEmpty()
 	req.Conditions.SetStatusCondition(conditionsv1.Condition{
 		Type:    hcov1beta1.ConditionReconcileComplete,
 		Status:  corev1.ConditionTrue,
@@ -586,33 +638,31 @@ func (r *ReconcileHyperConverged) aggregateComponentConditions(req *common.HcoRe
 		Message: reconcileCompletedMessage,
 	})
 
-	if _, conditionFound := req.Conditions[conditionsv1.ConditionDegraded]; conditionFound { // (#chart 1)
-		if _, conditionFound = req.Conditions[conditionsv1.ConditionProgressing]; !conditionFound { // (#chart 2)
-			req.Conditions.SetStatusCondition(conditionsv1.Condition{ // (#chart 3)
-				Type:    conditionsv1.ConditionProgressing,
-				Status:  corev1.ConditionFalse,
-				Reason:  reconcileCompleted,
-				Message: reconcileCompletedMessage,
-			})
-		} // else - Progressing is already exists
+	if req.Conditions.HasCondition(conditionsv1.ConditionDegraded) { // (#chart 1)
+
+		req.Conditions.SetStatusConditionIfUnset(conditionsv1.Condition{ // (#chart 2,3)
+			Type:    conditionsv1.ConditionProgressing,
+			Status:  corev1.ConditionFalse,
+			Reason:  reconcileCompleted,
+			Message: reconcileCompletedMessage,
+		})
+
+		req.Conditions.SetStatusConditionIfUnset(conditionsv1.Condition{ // (#chart 4,5)
+			Type:    conditionsv1.ConditionUpgradeable,
+			Status:  corev1.ConditionFalse,
+			Reason:  commonDegradedReason,
+			Message: "HCO is not Upgradeable due to degraded components",
+		})
+
+		req.Conditions.SetStatusConditionIfUnset(conditionsv1.Condition{ // (#chart 6,7)
+			Type:    conditionsv1.ConditionAvailable,
+			Status:  corev1.ConditionFalse,
+			Reason:  commonDegradedReason,
+			Message: "HCO is not available due to degraded components",
+		})
 
-		if _, conditionFound = req.Conditions[conditionsv1.ConditionUpgradeable]; !conditionFound { // (#chart 4)
-			req.Conditions.SetStatusCondition(conditionsv1.Condition{ // (#chart 5)
-				Type:    conditionsv1.ConditionUpgradeable,
-				Status:  corev1.ConditionFalse,
-				Reason:  commonDegradedReason,
-				Message: "HCO is not Upgradeable due to degraded components",
-			})
-		} // else - Upgradeable is already exists
-		if _, conditionFound = req.Conditions[conditionsv1.ConditionAvailable]; !conditionFound { // (#chart 6)
-			req.Conditions.SetStatusCondition(conditionsv1.Condition{ // (#chart 7)
-				Type:    conditionsv1.ConditionAvailable,
-				Status:  corev1.ConditionFalse,
-				Reason:  commonDegradedReason,
-				Message: "HCO is not available due to degraded components",
-			})
-		} // else - Available is already exists
 	} else {
+
 		// Degraded is not found. add it.
 		req.Conditions.SetStatusCondition(conditionsv1.Condition{ // (#chart 8)
 			Type:    conditionsv1.ConditionDegraded,
@@ -621,26 +671,24 @@ func (r *ReconcileHyperConverged) aggregateComponentConditions(req *common.HcoRe
 			Message: reconcileCompletedMessage,
 		})
 
-		if _, conditionFound = req.Conditions[conditionsv1.ConditionProgressing]; conditionFound { // (#chart 9)
+		if req.Conditions.HasCondition(conditionsv1.ConditionProgressing) { // (#chart 9)
 
-			if _, conditionFound = req.Conditions[conditionsv1.ConditionUpgradeable]; !conditionFound { // (#chart 10)
-				req.Conditions.SetStatusCondition(conditionsv1.Condition{ // (#chart 11)
-					Type:    conditionsv1.ConditionUpgradeable,
-					Status:  corev1.ConditionFalse,
-					Reason:  commonProgressingReason,
-					Message: "HCO is not Upgradeable due to progressing components",
-				})
-			} // else - Upgradeable is already exists
+			req.Conditions.SetStatusConditionIfUnset(conditionsv1.Condition{ // (#chart 10,11)
+				Type:    conditionsv1.ConditionUpgradeable,
+				Status:  corev1.ConditionFalse,
+				Reason:  commonProgressingReason,
+				Message: "HCO is not Upgradeable due to progressing components",
+			})
+
+			req.Conditions.SetStatusConditionIfUnset(conditionsv1.Condition{ // (#chart 12,13)
+				Type:    conditionsv1.ConditionAvailable,
+				Status:  corev1.ConditionTrue,
+				Reason:  reconcileCompleted,
+				Message: reconcileCompletedMessage,
+			})
 
-			if _, conditionFound = req.Conditions[conditionsv1.ConditionAvailable]; !conditionFound { // (#chart 12)
-				req.Conditions.SetStatusCondition(conditionsv1.Condition{ // (#chart 13)
-					Type:    conditionsv1.ConditionAvailable,
-					Status:  corev1.ConditionTrue,
-					Reason:  reconcileCompleted,
-					Message: reconcileCompletedMessage,
-				})
-			} // else - Available is already exists
 		} else {
+
 			req.Conditions.SetStatusCondition(conditionsv1.Condition{ // (#chart 14)
 				Type:    conditionsv1.ConditionProgressing,
 				Status:  corev1.ConditionFalse,
@@ -648,25 +696,23 @@ func (r *ReconcileHyperConverged) aggregateComponentConditions(req *common.HcoRe
 				Message: reconcileCompletedMessage,
 			})
 
-			if _, conditionFound = req.Conditions[conditionsv1.ConditionUpgradeable]; !conditionFound { // (#chart 15)
-				req.Conditions.SetStatusCondition(conditionsv1.Condition{ // (#chart 16)
-					Type:    conditionsv1.ConditionUpgradeable,
-					Status:  corev1.ConditionTrue,
-					Reason:  reconcileCompleted,
-					Message: reconcileCompletedMessage,
-				})
-			}
+			req.Conditions.SetStatusConditionIfUnset(conditionsv1.Condition{ // (#chart 15,16)
+				Type:    conditionsv1.ConditionUpgradeable,
+				Status:  corev1.ConditionTrue,
+				Reason:  reconcileCompleted,
+				Message: reconcileCompletedMessage,
+			})
+
+			req.Conditions.SetStatusConditionIfUnset(conditionsv1.Condition{ // (#chart 17,18)
+				Type:    conditionsv1.ConditionAvailable,
+				Status:  corev1.ConditionTrue,
+				Reason:  reconcileCompleted,
+				Message: reconcileCompletedMessage,
+			})
 
-			if _, conditionFound = req.Conditions[conditionsv1.ConditionAvailable]; !conditionFound { // (#chart 17) {
-				req.Conditions.SetStatusCondition(conditionsv1.Condition{ // (#chart 18)
-					Type:    conditionsv1.ConditionAvailable,
-					Status:  corev1.ConditionTrue,
-					Reason:  reconcileCompleted,
-					Message: reconcileCompletedMessage,
-				})
-			}
 		}
 	}
+
 	return allComponentsAreUp
 }
 
@@ -796,8 +842,51 @@ func (r *ReconcileHyperConverged) detectTaintedConfiguration(req *common.HcoRequ
 	}
 }
 
+func (r *ReconcileHyperConverged) firstLoopInitialization(request *common.HcoRequest) {
+	// Reload eventEmitter.
+	// The client should now find all the required resources.
+	r.eventEmitter.UpdateClient(request.Ctx, r.client, request.Logger)
+
+	// Initialize operand handler.
+	r.operandHandler.FirstUseInitiation(r.scheme, hcoutil.GetClusterInfo().IsOpenshift(), request.Instance)
+
+	// Avoid re-initializing.
+	r.firstLoop = false
+}
+
+// recoverHCOVersion recovers Spec.Version if upgrade missed when upgrade completed
+func (r *ReconcileHyperConverged) recoverHCOVersion(request *common.HcoRequest) {
+	knownHcoVersion, versionFound := request.Instance.Status.GetVersion(hcoVersionName)
+
+	if !r.upgradeMode &&
+		versionFound &&
+		(knownHcoVersion == r.ownVersion) &&
+		(request.Instance.Spec.Version != r.ownVersion) {
+
+		request.Instance.Spec.Version = r.ownVersion
+		request.Dirty = true
+	}
+}
+
+// This function performs migrations before starting the upgrade process
+// return true if the HyperConverged CR was modified; else, return false
+//
+// If the kubevirt-config configMap exists:
+// 1. create a backup of the configMap, if not exists
+// 2. if the configMap includes the live migration configurations, and they are not match to the default values,
+//    update the HyperConverged CR
+// 3. remove the kubevirt-config configMap
+// 4. return true if the CR was modified in #2
+const (
+	kvCmName         = "kubevirt-config"
+	backupKvCmName   = kvCmName + "-backup"
+	imsCmName        = "v2v-vmware"
+	liveMigrationKey = "migrations"
+	vddkInitImakeKey = "vddk-init-image"
+	crdName          = "hyperconvergeds.hco.kubevirt.io"
+)
+
 func (r *ReconcileHyperConverged) updateCrdStoredVersions(req *common.HcoRequest) (bool, error) {
-	const crdName = "hyperconvergeds.hco.kubevirt.io"
 	versionsToBeRemoved := []string{"v1alpha1"}
 
 	found := &apiextensionsv1.CustomResourceDefinition{}
@@ -833,8 +922,241 @@ func (r *ReconcileHyperConverged) updateCrdStoredVersions(req *common.HcoRequest
 	return false, nil
 }
 
-// getHyperconverged returns the name/namespace of the HyperConverged resource
-func getHyperconverged() (types.NamespacedName, error) {
+func (r *ReconcileHyperConverged) migrateBeforeUpgrade(req *common.HcoRequest) (bool, error) {
+	kvConfigModified, err := r.migrateKvConfigurations(req)
+	if err != nil {
+		return false, err
+	}
+
+	cdiConfigModified, err := r.migrateCdiConfigurations(req)
+	if err != nil {
+		return false, err
+	}
+
+	imsConfigModified, err := r.migrateImsConfigurations(req)
+	if err != nil {
+		return false, err
+	}
+
+	return kvConfigModified || cdiConfigModified || imsConfigModified, nil
+}
+
+func (r ReconcileHyperConverged) migrateKvConfigurations(req *common.HcoRequest) (bool, error) {
+	cm, err := r.getCm(kvCmName, req)
+	if err != nil {
+		return false, err
+	} else if cm == nil {
+		return false, nil
+	}
+
+	if err = r.makeCmBackup(cm, backupKvCmName, req); err != nil {
+		return false, err
+	}
+
+	modified := adoptOldKvConfigs(req, cm)
+
+	if !modified {
+		err = r.removeConfigMap(req, cm, kvCmName)
+		if err != nil {
+			return false, err
+		}
+	}
+
+	return modified, nil
+}
+
+func (r ReconcileHyperConverged) migrateCdiConfigurations(req *common.HcoRequest) (bool, error) {
+	cdi := operands.NewCDIWithNameOnly(req.Instance)
+
+	err := hcoutil.GetRuntimeObject(req.Ctx, r.client, cdi, req.Logger)
+	if err != nil {
+		if apierrors.IsNotFound(err) {
+			return false, nil
+		}
+		return false, err
+	}
+
+	return adoptCdiConfigs(req, cdi.Spec.Config), nil
+}
+
+func (r ReconcileHyperConverged) migrateImsConfigurations(req *common.HcoRequest) (bool, error) {
+	req.Logger.Info("read IMS configmap")
+	cm := &corev1.ConfigMap{
+		ObjectMeta: metav1.ObjectMeta{
+			Name:      imsCmName,
+			Namespace: req.Namespace,
+		},
+	}
+
+	if err := hcoutil.GetRuntimeObject(req.Ctx, r.client, cm, req.Logger); err != nil {
+		if apierrors.IsNotFound(err) {
+			req.Logger.Info("IMS configmap already removed")
+			return false, nil
+		}
+		req.Logger.Info("failed to get IMS configmap", "error", err.Error())
+		return false, err
+	}
+
+	modified := false
+	vddkInitImage, ok := cm.Data[vddkInitImakeKey]
+	if ok {
+		if req.Instance.Spec.VddkInitImage == nil {
+			req.Logger.Info("updating the HyperConverged CR from the IMS configMap")
+			req.Instance.Spec.VddkInitImage = &vddkInitImage
+			req.Dirty = true
+			modified = true
+		}
+	}
+
+	return modified, nil
+}
+
+func (r *ReconcileHyperConverged) removeConfigMap(req *common.HcoRequest, cm *corev1.ConfigMap, cmName string) error {
+	req.Logger.Info("removing the kubevirt configMap")
+	err := hcoutil.ComponentResourceRemoval(req.Ctx, r.client, cm, req.Name, req.Logger, false, true)
+	if err != nil {
+		return err
+	}
+
+	r.eventEmitter.EmitEvent(req.Instance, corev1.EventTypeNormal, "Killing", fmt.Sprintf("Removed ConfigMap %s", cmName))
+
+	refs := make([]corev1.ObjectReference, 0, len(req.Instance.Status.RelatedObjects))
+	for _, obj := range req.Instance.Status.RelatedObjects {
+		if obj.Kind == "ConfigMap" && obj.Name == cmName {
+			continue
+		}
+		refs = append(refs, obj)
+	}
+
+	req.Instance.Status.RelatedObjects = refs
+	req.StatusDirty = true
+
+	return nil
+}
+
+func (r *ReconcileHyperConverged) getCm(cmName string, req *common.HcoRequest) (*corev1.ConfigMap, error) {
+	cm := &corev1.ConfigMap{
+		ObjectMeta: metav1.ObjectMeta{
+			Name:      cmName,
+			Namespace: req.Namespace,
+		},
+	}
+
+	if err := hcoutil.GetRuntimeObject(req.Ctx, r.client, cm, req.Logger); err != nil {
+		if apierrors.IsNotFound(err) {
+			req.Logger.Info(fmt.Sprintf("%s configmap already removed", cmName))
+			return nil, nil
+		}
+		req.Logger.Info(fmt.Sprintf("failed to get %s configmap", cmName), "error", err.Error())
+		return nil, err
+	}
+
+	return cm, nil
+}
+
+func (r *ReconcileHyperConverged) makeCmBackup(cm *corev1.ConfigMap, backupName string, req *common.HcoRequest) error {
+	backupCm := &corev1.ConfigMap{
+		ObjectMeta: metav1.ObjectMeta{
+			Name:      backupName,
+			Namespace: cm.Namespace,
+			Labels:    cm.Labels,
+		},
+		Data: cm.Data,
+	}
+
+	req.Logger.Info(fmt.Sprintf("creating %s configmap backup", backupName))
+	if err := r.client.Create(req.Ctx, backupCm); err != nil {
+		if apierrors.IsAlreadyExists(err) {
+			req.Logger.Info(fmt.Sprintf("%s configmap backup already exists", backupName))
+		} else {
+			req.Logger.Info(fmt.Sprintf("failed to create %s configmap backup", backupName), "error", err.Error())
+			return err
+		}
+	} else {
+		r.eventEmitter.EmitEvent(req.Instance, corev1.EventTypeNormal, "Created", fmt.Sprintf("Created ConfigMap %s", backupName))
+	}
+
+	return nil
+}
+
+// Read the old KubeVit configuration from the config map, and move them to the HyperConverged CR
+//
+// In case of wrong foramt of the configmap, the HCO ignores this error (but print it to the log) in order to prevent
+// an infinite loop (returning error will cause the same error again and again, and the only way to stop the loop
+// is to manually fix or delete the wrong configMap).
+func adoptOldKvConfigs(req *common.HcoRequest, cm *corev1.ConfigMap) bool {
+	modified := false
+	kvLiveMigrationConfig, ok := cm.Data[liveMigrationKey]
+	if !ok {
+		return false
+	}
+	kvCmLiveMigrationConfig := hcov1beta1.LiveMigrationConfigurations{}
+	req.Instance.Spec.LiveMigrationConfig.DeepCopyInto(&kvCmLiveMigrationConfig)
+	err := yaml.NewYAMLOrJSONDecoder(strings.NewReader(kvLiveMigrationConfig), 1024).Decode(&kvCmLiveMigrationConfig)
+	if err != nil {
+		req.Logger.Error(err, "Failed to read the KubeVirt ConfigMap, and its content was ignored. This ConfigMap will be deleted. The backup ConfigMap called "+backupKvCmName)
+		return false
+	}
+
+	if !reflect.DeepEqual(req.Instance.Spec.LiveMigrationConfig, kvCmLiveMigrationConfig) {
+		req.Logger.Info("updating the HyperConverged CR from the KubeVirt configMap")
+		kvConfigMapToHyperConvergedCr(req, kvCmLiveMigrationConfig)
+
+		modified = true
+		req.Dirty = true
+	}
+
+	return modified
+}
+
+func adoptCdiConfigs(req *common.HcoRequest, cdiCfg *cdiv1beta1.CDIConfigSpec) bool {
+	modified := false
+	if cdiCfg != nil {
+		if req.Instance.Spec.ScratchSpaceStorageClass == nil && cdiCfg.ScratchSpaceStorageClass != nil {
+			req.Instance.Spec.ScratchSpaceStorageClass = new(string)
+			*req.Instance.Spec.ScratchSpaceStorageClass = *cdiCfg.ScratchSpaceStorageClass
+			modified = true
+		}
+
+		if cdiCfg.PodResourceRequirements != nil {
+			if req.Instance.Spec.ResourceRequirements == nil {
+				req.Instance.Spec.ResourceRequirements = &hcov1beta1.OperandResourceRequirements{}
+			}
+
+			if req.Instance.Spec.ResourceRequirements.StorageWorkloads == nil {
+				req.Instance.Spec.ResourceRequirements.StorageWorkloads = cdiCfg.PodResourceRequirements.DeepCopy()
+				modified = true
+			}
+		}
+	}
+
+	if modified {
+		req.Dirty = true
+	}
+
+	return modified
+}
+
+func kvConfigMapToHyperConvergedCr(req *common.HcoRequest, kvCmLMConfig hcov1beta1.LiveMigrationConfigurations) {
+	if kvCmLMConfig.BandwidthPerMigration != nil {
+		req.Instance.Spec.LiveMigrationConfig.BandwidthPerMigration = kvCmLMConfig.BandwidthPerMigration
+	}
+	if kvCmLMConfig.CompletionTimeoutPerGiB != nil {
+		req.Instance.Spec.LiveMigrationConfig.CompletionTimeoutPerGiB = kvCmLMConfig.CompletionTimeoutPerGiB
+	}
+	if kvCmLMConfig.ParallelMigrationsPerCluster != nil {
+		req.Instance.Spec.LiveMigrationConfig.ParallelMigrationsPerCluster = kvCmLMConfig.ParallelMigrationsPerCluster
+	}
+	if kvCmLMConfig.ParallelOutboundMigrationsPerNode != nil {
+		req.Instance.Spec.LiveMigrationConfig.ParallelOutboundMigrationsPerNode = kvCmLMConfig.ParallelOutboundMigrationsPerNode
+	}
+	if kvCmLMConfig.ProgressTimeout != nil {
+		req.Instance.Spec.LiveMigrationConfig.ProgressTimeout = kvCmLMConfig.ProgressTimeout
+	}
+}
+
+// getHyperConvergedNamespacedName returns the name/namespace of the HyperConverged resource
+func getHyperConvergedNamespacedName() (types.NamespacedName, error) {
 	hco := types.NamespacedName{
 		Name: hcoutil.HyperConvergedName,
 	}
@@ -865,15 +1187,6 @@ func getSecondaryCRPlaceholder() (types.NamespacedName, error) {
 	return hco, nil
 }
 
-func contains(slice []string, s string) bool {
-	for _, element := range slice {
-		if element == s {
-			return true
-		}
-	}
-	return false
-}
-
 func drop(slice []string, s string) ([]string, bool) {
 	newSlice := []string{}
 	dropped := false
@@ -890,3 +1203,23 @@ func drop(slice []string, s string) ([]string, bool) {
 func init() {
 	randomConstSuffix = uuid.New().String()
 }
+
+func checkFinalizers(req *common.HcoRequest) bool {
+	finDropped := false
+
+	if hcoutil.ContainsString(req.Instance.ObjectMeta.Finalizers, badFinalizerName) {
+		req.Logger.Info("removing a finalizer set in the past (without a fully qualified name)")
+		req.Instance.ObjectMeta.Finalizers, finDropped = drop(req.Instance.ObjectMeta.Finalizers, badFinalizerName)
+		req.Dirty = req.Dirty || finDropped
+	}
+	if req.Instance.ObjectMeta.DeletionTimestamp.IsZero() {
+		// Add the finalizer if it's not there
+		if !hcoutil.ContainsString(req.Instance.ObjectMeta.Finalizers, FinalizerName) {
+			req.Logger.Info("setting a finalizer (with fully qualified name)")
+			req.Instance.ObjectMeta.Finalizers = append(req.Instance.ObjectMeta.Finalizers, FinalizerName)
+			req.Dirty = req.Dirty || finDropped
+		}
+		return true
+	}
+	return false
+}
