@@ -108,6 +108,15 @@ var _ = Describe("SSP Operands", func() {
 					foundResource),
 			).To(BeNil())
 			Expect(foundResource.Spec).To(Equal(expectedResource.Spec))
+
+			// ObjectReference should have been updated
+			Expect(hco.Status.RelatedObjects).To(Not(BeNil()))
+			objectRefOutdated, err := reference.GetReference(handler.Scheme, existingResource)
+			Expect(err).To(BeNil())
+			objectRefFound, err := reference.GetReference(handler.Scheme, foundResource)
+			Expect(err).To(BeNil())
+			Expect(hco.Status.RelatedObjects).To(Not(ContainElement(*objectRefOutdated)))
+			Expect(hco.Status.RelatedObjects).To(ContainElement(*objectRefFound))
 		})
 
 		Context("Node placement", func() {
@@ -369,10 +378,10 @@ var _ = Describe("SSP Operands", func() {
 				// Initialize RelatedObjects with a bunch of objects
 				// including old SSP ones.
 				for _, objRef := range oldSSPRelatedObjects() {
-					v1.SetObjectReference(&hco.Status.RelatedObjects, objRef)
+					Expect(v1.SetObjectReference(&hco.Status.RelatedObjects, objRef)).ToNot(HaveOccurred())
 				}
 				for _, objRef := range otherRelatedObjects() {
-					v1.SetObjectReference(&hco.Status.RelatedObjects, objRef)
+					Expect(v1.SetObjectReference(&hco.Status.RelatedObjects, objRef)).ToNot(HaveOccurred())
 				}
 
 				cl := commonTestUtils.InitClient(nil)
@@ -398,10 +407,10 @@ var _ = Describe("SSP Operands", func() {
 				// Initialize RelatedObjects with a bunch of objects
 				// including old SSP ones.
 				for _, objRef := range oldSSPRelatedObjects() {
-					v1.SetObjectReference(&hco.Status.RelatedObjects, objRef)
+					Expect(v1.SetObjectReference(&hco.Status.RelatedObjects, objRef)).ToNot(HaveOccurred())
 				}
 				for _, objRef := range otherRelatedObjects() {
-					v1.SetObjectReference(&hco.Status.RelatedObjects, objRef)
+					Expect(v1.SetObjectReference(&hco.Status.RelatedObjects, objRef)).ToNot(HaveOccurred())
 				}
 
 				cl := commonTestUtils.InitClient(nil)
@@ -417,7 +426,7 @@ var _ = Describe("SSP Operands", func() {
 				// Now simulate "status update failure",
 				// i.e. related objects aren't removed.
 				for _, objRef := range oldSSPRelatedObjects() {
-					v1.SetObjectReference(&hco.Status.RelatedObjects, objRef)
+					Expect(v1.SetObjectReference(&hco.Status.RelatedObjects, objRef)).ToNot(HaveOccurred())
 				}
 
 				// Simulate another reconciliation cycle
@@ -436,6 +445,55 @@ var _ = Describe("SSP Operands", func() {
 				}
 			})
 		})
+
+		Context("Cache", func() {
+			cl := commonTestUtils.InitClient([]runtime.Object{})
+			handler := newSspHandler(cl, commonTestUtils.GetScheme())
+
+			It("should start with empty cache", func() {
+				Expect(handler.hooks.(*sspHooks).cache).To(BeNil())
+			})
+
+			It("should update the cache when reading full CR", func() {
+				cr, err := handler.hooks.getFullCr(hco)
+				Expect(err).ToNot(HaveOccurred())
+				Expect(cr).ToNot(BeNil())
+				Expect(handler.hooks.(*sspHooks).cache).ToNot(BeNil())
+
+				By("compare pointers to make sure cache is working", func() {
+					Expect(handler.hooks.(*sspHooks).cache == cr).Should(BeTrue())
+
+					cdi1, err := handler.hooks.getFullCr(hco)
+					Expect(err).ToNot(HaveOccurred())
+					Expect(cdi1).ToNot(BeNil())
+					Expect(cr == cdi1).Should(BeTrue())
+				})
+			})
+
+			It("should remove the cache on reset", func() {
+				handler.hooks.(*sspHooks).reset()
+				Expect(handler.hooks.(*sspHooks).cache).To(BeNil())
+			})
+
+			It("check that reset actually cause creating of a new cached instance", func() {
+				crI, err := handler.hooks.getFullCr(hco)
+				Expect(err).ToNot(HaveOccurred())
+				Expect(crI).ToNot(BeNil())
+				Expect(handler.hooks.(*sspHooks).cache).ToNot(BeNil())
+
+				handler.hooks.(*sspHooks).reset()
+				Expect(handler.hooks.(*sspHooks).cache).To(BeNil())
+
+				crII, err := handler.hooks.getFullCr(hco)
+				Expect(err).ToNot(HaveOccurred())
+				Expect(crII).ToNot(BeNil())
+				Expect(handler.hooks.(*sspHooks).cache).ToNot(BeNil())
+
+				Expect(crI == crII).To(BeFalse())
+				Expect(handler.hooks.(*sspHooks).cache == crI).To(BeFalse())
+				Expect(handler.hooks.(*sspHooks).cache == crII).To(BeTrue())
+			})
+		})
 	})
 })
 
